//-----------------------------------------------------------------------------
// File: Meshes.cpp
//
// Desc: For advanced geometry, most apps will prefer to load pre-authored
//       Meshes from a file. Fortunately, when using Meshes, D3DX does most of
//       the work for this, parsing a geometry file and creating vertx buffers
//       (and index buffers) for us. This tutorial shows how to use a D3DXMESH
//       object, including loading it from a file and rendering it. One thing
//       D3DX does not handle for us is the materials and textures for a mesh,
//       so note that we have to handle those manually.
//
//       Note: one advanced (but nice) feature that we don't show here is that
//       when cloning a mesh we can specify the FVF. So, regardless of how the
//       mesh was authored, we can add/remove normals, add more texture
//       coordinate sets (for multi-texturing), etc.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
#include <Windows.h>
#include <mmsystem.h>
#include <d3dx9.h>
#pragma warning( disable : 4996 ) // disable deprecated warning 
#include <strsafe.h>
#pragma warning( default : 4996 ) 

// _ݒ
struct VERTEX_XYZ_RHW_COLOR_UV
{
	float x, y, z;
	float rhw;
	D3DCOLOR color;
	float u, v;
};

#define	D3DFVF_XYZ_RHW_COLOR_UV	( D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1 )

struct VERTEX_XYZ_COLOR_UV
{
	float x, y, z;
	D3DCOLOR color;
	float u, v;
};

#define	D3DFVF_XYZ_COLOR_UV ( D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1 )

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
LPDIRECT3D9             g_pD3D           = NULL; // Used to create the D3DDevice
LPDIRECT3DDEVICE9       g_pd3dDevice     = NULL; // Our rendering device

LPD3DXMESH              g_pMesh          = NULL; // Our mesh object in sysmem
D3DMATERIAL9*           g_pMeshMaterials = NULL; // Materials for our mesh
LPDIRECT3DTEXTURE9*     g_pMeshTextures  = NULL; // Textures for our mesh
DWORD                   g_dwNumMaterials = 0L;   // Number of mesh materials

LPDIRECT3DSURFACE9 g_pDirect3DSurfaceBack = NULL;
LPDIRECT3DSURFACE9 g_pZStencilSurface = NULL;
LPDIRECT3DTEXTURE9 g_pD3DTexture[2] = { NULL, NULL };
LPDIRECT3DSURFACE9 g_pD3DSurface[2] = { NULL, NULL };

LPDIRECT3DTEXTURE9 g_pD3DTexBack = NULL;

LPD3DXEFFECT g_pEffect = NULL;

//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
HRESULT InitD3D( HWND hWnd )
{
    // Create the D3D object.
    if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
        return E_FAIL;

    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_X8R8G8B8;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	d3dpp.BackBufferWidth = 512;
	d3dpp.BackBufferHeight = 512;
	d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	d3dpp.Flags = D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;

    // Create the D3DDevice
    if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &d3dpp, &g_pd3dDevice ) ) )
    {
        return E_FAIL;
    }

    // Turn on the zbuffer
    g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );

    // Turn on ambient lighting 
    g_pd3dDevice->SetRenderState( D3DRS_AMBIENT, 0xffffffff );

	// eT[tF[X
	g_pd3dDevice->GetRenderTarget( 0, &g_pDirect3DSurfaceBack );

	g_pd3dDevice->CreateDepthStencilSurface( 512, 512,
		D3DFMT_D16, D3DMULTISAMPLE_NONE, 0, FALSE, &g_pZStencilSurface, NULL );

	D3DSURFACE_DESC Desc;
	g_pDirect3DSurfaceBack->GetDesc( &Desc );

	g_pd3dDevice->CreateTexture( 512, 512, 1, D3DUSAGE_RENDERTARGET,
		Desc.Format, D3DPOOL_DEFAULT, &g_pD3DTexture[0], NULL );
	g_pD3DTexture[0]->GetSurfaceLevel( 0, &g_pD3DSurface[0] );

	g_pd3dDevice->CreateTexture( 128, 128, 1, D3DUSAGE_DYNAMIC,
		D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &g_pD3DTexture[1], NULL );

	// VF[_ǂݍ
	LPD3DXBUFFER pCompilationErrors;
    if( FAILED ( 
		D3DXCreateEffectFromFile( g_pd3dDevice, "hlsl.fx", NULL, NULL, D3DXSHADER_DEBUG/* | D3DXSHADER_SKIPOPTIMIZATION*/, NULL, &g_pEffect, &pCompilationErrors ) )
	){
		MessageBox( NULL, (LPCSTR)pCompilationErrors->GetBufferPointer(), "Error", MB_OK );
        return E_FAIL;
	}

	g_pEffect->OnResetDevice();

	// eNX`ǂݍ
	D3DXCreateTextureFromFile( g_pd3dDevice, "lobbyzpos.JPG", &g_pD3DTexBack );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: InitGeometry()
// Desc: Load the mesh and build the material and texture arrays
//-----------------------------------------------------------------------------
HRESULT InitGeometry()
{
    LPD3DXBUFFER pD3DXMtrlBuffer;

    // Load the mesh from the specified file
    if( FAILED( D3DXLoadMeshFromX( "Tiger.x", D3DXMESH_SYSTEMMEM, 
                                   g_pd3dDevice, NULL, 
                                   &pD3DXMtrlBuffer, NULL, &g_dwNumMaterials, 
                                   &g_pMesh ) ) )
    {
        // If model is not in current folder, try parent folder
        if( FAILED( D3DXLoadMeshFromX( "..\\Tiger.x", D3DXMESH_SYSTEMMEM, 
                                    g_pd3dDevice, NULL, 
                                    &pD3DXMtrlBuffer, NULL, &g_dwNumMaterials, 
                                    &g_pMesh ) ) )
        {
            MessageBox(NULL, "Could not find tiger.x", "Meshes.exe", MB_OK);
            return E_FAIL;
        }
    }

    // We need to extract the material properties and texture names from the 
    // pD3DXMtrlBuffer
    D3DXMATERIAL* d3dxMaterials = (D3DXMATERIAL*)pD3DXMtrlBuffer->GetBufferPointer();
    g_pMeshMaterials = new D3DMATERIAL9[g_dwNumMaterials];
    if( g_pMeshMaterials == NULL )
        return E_OUTOFMEMORY;
    g_pMeshTextures  = new LPDIRECT3DTEXTURE9[g_dwNumMaterials];
    if( g_pMeshTextures == NULL )
        return E_OUTOFMEMORY;

    for( DWORD i=0; i<g_dwNumMaterials; i++ )
    {
        // Copy the material
        g_pMeshMaterials[i] = d3dxMaterials[i].MatD3D;

        // Set the ambient color for the material (D3DX does not do this)
        g_pMeshMaterials[i].Ambient = g_pMeshMaterials[i].Diffuse;

        g_pMeshTextures[i] = NULL;
        if( d3dxMaterials[i].pTextureFilename != NULL && 
            lstrlen(d3dxMaterials[i].pTextureFilename) > 0 )
        {
            // Create the texture
            if( FAILED( D3DXCreateTextureFromFile( g_pd3dDevice, 
                                                d3dxMaterials[i].pTextureFilename, 
                                                &g_pMeshTextures[i] ) ) )
            {
                // If texture is not in current folder, try parent folder
                const TCHAR* strPrefix = TEXT("..\\");
                TCHAR strTexture[MAX_PATH];
                StringCchCopy( strTexture, MAX_PATH, strPrefix );
                StringCchCat( strTexture, MAX_PATH, d3dxMaterials[i].pTextureFilename );
                // If texture is not in current folder, try parent folder
                if( FAILED( D3DXCreateTextureFromFile( g_pd3dDevice, 
                                                    strTexture, 
                                                    &g_pMeshTextures[i] ) ) )
                {
                    MessageBox(NULL, "Could not find texture map", "Meshes.exe", MB_OK);
                }
            }
        }
    }

    // Done with the material buffer
    pD3DXMtrlBuffer->Release();

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
VOID Cleanup()
{
	if( g_pZStencilSurface ) g_pZStencilSurface->Release();
	if( g_pEffect ) g_pEffect->Release();

	if( g_pD3DTexBack ) g_pD3DTexBack->Release();

	for( int i = 0; i < 2; ++i ){
		if( g_pD3DSurface[i] ) g_pD3DSurface[i]->Release();
		if( g_pD3DTexture[i] ) g_pD3DTexture[i]->Release();
	}

	g_pDirect3DSurfaceBack->Release();

    if( g_pMeshMaterials != NULL ) 
        delete[] g_pMeshMaterials;

    if( g_pMeshTextures )
    {
        for( DWORD i = 0; i < g_dwNumMaterials; i++ )
        {
            if( g_pMeshTextures[i] )
                g_pMeshTextures[i]->Release();
        }
        delete[] g_pMeshTextures;
    }
    if( g_pMesh != NULL )
        g_pMesh->Release();
    
    if( g_pd3dDevice != NULL )
        g_pd3dDevice->Release();

    if( g_pD3D != NULL )
        g_pD3D->Release();
}



//-----------------------------------------------------------------------------
// Name: SetupMatrices()
// Desc: Sets up the world, view, and projection transform matrices.
//-----------------------------------------------------------------------------
VOID SetupMatrices()
{
    // Set up world matrix
    D3DXMATRIXA16 matWorld;
    D3DXMatrixRotationY( &matWorld, timeGetTime()/1000.0f );
    g_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

    // Set up our view matrix. A view matrix can be defined given an eye point,
    // a point to lookat, and a direction for which way is up. Here, we set the
    // eye five units back along the z-axis and up three units, look at the 
    // origin, and define "up" to be in the y-direction.
    D3DXVECTOR3 vEyePt( 0.0f, 3.0f,-5.0f );
    D3DXVECTOR3 vLookatPt( 0.0f, 0.0f, 0.0f );
    D3DXVECTOR3 vUpVec( 0.0f, 1.0f, 0.0f );
    D3DXMATRIXA16 matView;
    D3DXMatrixLookAtLH( &matView, &vEyePt, &vLookatPt, &vUpVec );
    g_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );

    // For the projection matrix, we set up a perspective transform (which
    // transforms geometry from 3D view space to 2D viewport space, with
    // a perspective divide making objects smaller in the distance). To build
    // a perpsective transform, we need the field of view (1/4 pi is common),
    // the aspect ratio, and the near and far clipping planes (which define at
    // what distances geometry should be no longer be rendered).
    D3DXMATRIXA16 matProj;
    D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, 1.0f, 1.0f, 100.0f );
    g_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );
}




//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
VOID Render()
{
	BYTE KeyBuffer[256];
	GetKeyboardState( KeyBuffer );
	static bool IsPush = false;
	static int EffectID = 0;

	if( KeyBuffer[VK_SPACE] & 0x80 ){
		if( !IsPush ){
			IsPush = true;
			++EffectID;
			EffectID %= 3;
			MessageBeep( -1 );
		}
	}
	else IsPush = false;

	// geNX`
	DWORD Time = timeGetTime();

	D3DLOCKED_RECT LockedRect;
	if( FAILED( g_pD3DTexture[1]->LockRect( 0, &LockedRect, NULL, D3DLOCK_DISCARD ) )) MessageBeep(-1);
	unsigned char *pTmp = new unsigned char[LockedRect.Pitch];

	for( int y = 0; y < 128; ++y ){
		for( int x = 0; x < 128; ++x ){
			float l = sqrtf( (float)( ( x - 64 ) * ( x - 64 ) + ( y - 64 ) * ( y - 64 ) ) );
			float ll = sinf( (float) l / 5 + fmodf( (float)Time / 100, 3.1415926535f * 2 ) );

			pTmp[x * 4 + 0] = (unsigned char)( ( ll * (float)( x - 64 ) / l + 1 ) * 127 );
			pTmp[x * 4 + 1] = (unsigned char)( ( ll * (float)( y - 64 ) / l + 1 ) * 127 );
		}
		memcpy( (unsigned char *)LockedRect.pBits + LockedRect.Pitch * y, pTmp, LockedRect.Pitch );
	}
	delete [] pTmp;

	g_pD3DTexture[1]->UnlockRect( 0 );

    // Begin the scene
    if( SUCCEEDED( g_pd3dDevice->BeginScene() ) )
    {
		g_pd3dDevice->SetRenderTarget( 0, g_pD3DSurface[0] );
		g_pd3dDevice->SetDepthStencilSurface( g_pZStencilSurface );

		g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
		g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
		g_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
		g_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
		g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, FALSE );
		g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, TRUE );

		// Clear the backbuffer and the zbuffer
		g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 
							 D3DCOLOR_XRGB(0,0,255), 1.0f, 0 );

		// wi_O
		VERTEX_XYZ_RHW_COLOR_UV Vertex[4] = {
			{   0,   0, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 0, 0 },
			{ 512,   0, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 1, 0 },
			{   0, 512, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 0, 1 },
			{ 512, 512, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 1, 1 }
		};

		g_pd3dDevice->SetTexture( 0, g_pD3DTexBack );
		g_pd3dDevice->SetFVF( D3DFVF_XYZ_RHW_COLOR_UV );
		g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, Vertex, sizeof( VERTEX_XYZ_RHW_COLOR_UV ) );


		g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );

        // Setup the world, view, and projection matrices
        SetupMatrices();

        // Meshes are divided into subsets, one for each material. Render them in
        // a loop
        for( DWORD i=0; i<g_dwNumMaterials; i++ )
        {
            // Set the material and texture for this subset
            g_pd3dDevice->SetMaterial( &g_pMeshMaterials[i] );
            g_pd3dDevice->SetTexture( 0, g_pMeshTextures[i] );
        
            // Draw the mesh subset
            g_pMesh->DrawSubset( i );
        }

		if( EffectID == 0 || EffectID == 1 ){
			D3DVIEWPORT9 Viewport = { 0, 0, 512, 512, 0, 1 };
			g_pd3dDevice->SetViewport( &Viewport );

			g_pd3dDevice->SetRenderTarget( 0, g_pDirect3DSurfaceBack );
			g_pd3dDevice->SetDepthStencilSurface( g_pZStencilSurface );

			g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
			g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, FALSE );
			g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );

			VERTEX_XYZ_COLOR_UV Vertex[4] = {
				{ -1, -1, 0, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 0, 1 },
				{ -1,  1, 0, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 0, 0 },
				{  1, -1, 0, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 1, 1 },
				{  1,  1, 0, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 1, 0 }
			};

			g_pd3dDevice->SetFVF( D3DFVF_XYZ_COLOR_UV );

			g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );
			g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
			g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );

			if( EffectID == 0 ){
				g_pEffect->SetTexture( "TexMap", g_pD3DTexture[0] );
				g_pEffect->SetTexture( "TexMap2", g_pD3DTexture[1] );
				g_pEffect->CommitChanges();
				g_pEffect->SetTechnique( g_pEffect->GetTechniqueByName( "RippleShader" ) );
			}

			if( EffectID == 1 ){
				g_pEffect->SetTexture( "TexMap", g_pD3DTexture[0] );
				g_pEffect->CommitChanges();
				g_pEffect->SetTechnique( g_pEffect->GetTechniqueByName( "BlurShader" ) );
			}

			UINT nPasses;
			g_pEffect->Begin( &nPasses, 0 );
			g_pEffect->BeginPass( 0 );
			g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, Vertex, sizeof( VERTEX_XYZ_COLOR_UV ) );
			g_pEffect->EndPass();

			g_pEffect->End();
		}

		if( EffectID == 2 ){
			g_pd3dDevice->SetRenderTarget( 0, g_pDirect3DSurfaceBack );
			VERTEX_XYZ_RHW_COLOR_UV Vertex[4] = {
				{   0,   0, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 0, 0 },
				{ 512,   0, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 1, 0 },
				{   0, 512, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 0, 1 },
				{ 512, 512, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 255, 255 ), 1, 1 }
			};

			// eNX`ݒ
			g_pd3dDevice->SetTexture( 0, g_pD3DTexture[0] );

			// _tH[}bg̐ݒ
			g_pd3dDevice->SetFVF( D3DFVF_XYZ_RHW_COLOR_UV );

			g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, Vertex, sizeof( VERTEX_XYZ_RHW_COLOR_UV ) );
		}

		HDC hDC;
		g_pDirect3DSurfaceBack->GetDC( &hDC );
		char Effect[][256] = { "Ripple", "Concentration Line", "None" };
		char Buffer[256];
		sprintf( Buffer, "Push Space Key [Effect: %s]", Effect[EffectID] );
		TextOut( hDC, 10, 10, Buffer, (int)strlen( Buffer ) );
		g_pDirect3DSurfaceBack->ReleaseDC( hDC );

        // End the scene
        g_pd3dDevice->EndScene();
    }

    // Present the backbuffer contents to the display
    g_pd3dDevice->Present( NULL, NULL, NULL, NULL );
}




//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
        case WM_DESTROY:
            Cleanup();
            PostQuitMessage( 0 );
            return 0;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}




//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    // Register the window class
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L, 
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      "D3D Tutorial", NULL };
    RegisterClassEx( &wc );

    // Create the application's window
    HWND hWnd = CreateWindow( "D3D Tutorial", "Post Effect", 
                              WS_OVERLAPPEDWINDOW, 100, 100, 512, 512,
                              NULL, NULL, wc.hInstance, NULL );

    // Initialize Direct3D
    if( SUCCEEDED( InitD3D( hWnd ) ) )
    { 
        // Create the scene geometry
        if( SUCCEEDED( InitGeometry() ) )
        {
            // Show the window
            ShowWindow( hWnd, SW_SHOWDEFAULT );
            UpdateWindow( hWnd );

            // Enter the message loop
            MSG msg; 
            ZeroMemory( &msg, sizeof(msg) );
            while( msg.message!=WM_QUIT )
            {
                if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
                {
                    TranslateMessage( &msg );
                    DispatchMessage( &msg );
                }
                else
                    Render();
            }
        }
    }

    UnregisterClass( "D3D Tutorial", wc.hInstance );
    return 0;
}



